server = function(input, output, session = session) {
  ## note that, within R shiny, the "server" function is basically
  ## a script with nested function definitions.


  ## Define overarching variable for this session
  df_plotted_points <- data.frame(lon = numeric(), lat=numeric())

  #### DEFINE REACTIVE EXPRESSIONS ####
  ## In sloppy language: reactive expressions are
  ## variable like functions which will change according to web user actions

  ####  "re_variable_of_interest": process click point on map ####
  re_variable_of_interest <- reactive(
    {
      sf_pnt <- st_point(c(input$click_coordinates[2], input$click_coordinates[1]))
      #browser()
      n_index_inters <- sf::st_intersects(sf_pnt, sf_mapVOI)[[1]]
      if(length(n_index_inters)>0)
        str_VOI <- sf_mapVOI$voi[ n_index_inters]
      else
        str_VOI <- str_Outside_Map
      return(str_VOI)
    })


  #### re_df_student_loc_and_var:  load saved student locations ####
  re_df_student_loc_and_var <- reactive({

    vs_datafiles <- list.files(path = "./student_data/" ,
                               pattern = ".csv",
                               full.names = TRUE)



    if(length(vs_datafiles)>0)
    {
      df_student_loc_and_var <- do.call(rbind,lapply(vs_datafiles,read.csv))
      df_student_loc_and_var <- rbind(df_student_loc_and_var,
                                      df_initial_loc_and_var)
    } else
    {
      ## The initial location (for example Wageningen campus)
      ## This also prevents to take into account that there
      ## might be no input at all.
      df_student_loc_and_var <- df_initial_loc_and_var
      #browser()
    }

    ## Update map & graphs every 2 seconds
    invalidateLater(2000, session)


    return(df_student_loc_and_var)

  })


  #### re_remove_student_data ####
  re_remove_student_data <- reactive(
    {
      bo_removed <- FALSE
      li_url_query <- parseQueryString(session$clientData$url_search)
      #browser()
      if(!is.null(li_url_query[["remove"]]))
      {
        return(TRUE)
      } else
      return(FALSE)
    })

  #### re_is_teacher_version : show teacher version of interface ####
  re_is_teacher_version <- reactive(
    {
      li_url_query <- parseQueryString(session$clientData$url_search)
      if(!is.null(li_url_query[["teacher"]]))
      {
        return(TRUE)
      } else
      return(FALSE)

    })

  #### DEFINE ELEMENTS ####

  #### Output "histogram_of_var"  ####
  output$histogram_of_var = renderPlot({

    ## Load data
    df_student_loc_and_var <- re_df_student_loc_and_var()

    ## check which points are not plotted yet
    df_student_loc_and_var_to_plot <-
      subset(df_student_loc_and_var, !( (lon %in% df_plotted_points$lon) &
                                          (lat %in% df_plotted_points$lat)
      )
      )


    if(nrow(df_student_loc_and_var_to_plot)>0)
    {

      df_plotted_points <<- rbind(df_plotted_points,
                                  df_student_loc_and_var[,c("lon", "lat")]
      )

      #### Output "map" add locations  ####
      leafletProxy("map") %>%  addCircles(map = .,
                                          lng = df_student_loc_and_var_to_plot$lon,
                                          lat = df_student_loc_and_var_to_plot$lat
      )
    }


    ## Create histogram
    df_student_loc_and_var$variable_of_interest <-
      factor(df_student_loc_and_var$variable_of_interest,
             levels = sort(c(unique(sf_mapVOI$voi)), #str_Outside_Map),  #unique(df_student_loc_and_var$variable_of_interest),
                           decreasing = TRUE ) # Decreasing, because the later plot is flipped
      )

    ## size of dataset to show
    n_to_show <- length(unique(df_student_loc_and_var$variable_of_interest))

    ## total possible to show
    n_total <- length(unique(sf_mapVOI$voi))

    levels(df_student_loc_and_var$variable_of_interest)

    vs_colors_tmp <- rev(vs_colors)
    vs_colors_tmp <- vs_colors_tmp[
      as.logical(table(df_student_loc_and_var[,"variable_of_interest"]))
    ]

    ggplot(df_student_loc_and_var ,
           aes(variable_of_interest) ) +
      geom_bar(fill = vs_colors_tmp, width=0.65) +
      scale_x_discrete(drop=FALSE) +
      coord_flip() +
      ylab(df_texts["n_of_locations", str_language]) +
      xlab("") +
      theme(axis.text = element_text(size = 20)) +
      scale_y_continuous(breaks = function(x) unique(floor(pretty(seq(0, (max(x) + 1) * 1.1)))))





  })


  ####   Output "map" ####
  output$map = renderLeaflet({
    #par(mar=c(1,1,1,1))
    mapview::mapview(sf_mapVOI,
                     zcol="voi",
                     #color = vs_colors,
                     col.regions = vs_colors,
                     popup=FALSE,
                     map.types = c("CartoDB.Positron", "OpenStreetMap","OpenTopoMap",  "Esri.WorldImagery"),
                     #maxpixels = 1000,
                     alpha.regions = 0.65, #
                     # Note: if one changes alpha.regions, also the legend colors as defined in mian.R have to be adjusted
                     pane = NULL,
                     label = FALSE,
                     canvas = FALSE,
                     layer.name = df_texts["layer_name", str_language]
    )@map %>%
      setView(lng = 5.4, lat = 52.35, zoom = 7) %>%
      addMapPane("ames_circles", zIndex = 420) %>% # shown above ames_lines
      onRender(  #mousemove null
        "function(el,x){
                    this.on('click', function(e) {
                        var lat = e.latlng.lat;
                        var lng = e.latlng.lng;
                        var coord = [lat, lng];
                        Shiny.onInputChange('click_coordinates', coord)
                    });
                    this.on('click', function(e) {
                        Shiny.onInputChange('click_coordinates', coord)
                    })
                }"
      )

  })




  ####  Output "location_and_var": add clicked point to map, invite to submit ####
  output$location_and_var <- renderText({
    if(is.null(input$click_coordinates)) {
      df_texts["first_instructions_zoom_and_click", str_language]
    } else {
      leafletProxy("map") %>%  removeShape(layerId = "student_location")
      leafletProxy("map") %>%      addCircles(map = .,
                                              layerId = "student_location",
                                              lng = input$click_coordinates[2],
                                              lat = input$click_coordinates[1])
      paste0("", format(round(input$click_coordinates[1], digits=4), nsmall=4), " ", df_texts["degrees_north_latitud", str_language],
             "\n ", format(round(input$click_coordinates[2], digits=4), nsmall=4), " ", df_texts["degrees_east_longitude", str_language],
             "\n", re_variable_of_interest(),
             "\n", df_texts["invite_to_submit_location", str_language]
      )
    }
  })




  #### Output "show_version": show the student names in the teacher projection ####
  output$show_version <- renderText({
    df_student_loc_and_var <- re_df_student_loc_and_var()


    vs_names <- df_student_loc_and_var$student_name

    vs_names <- vs_names[nzchar(vs_names)]
    vs_names <- vs_names[!is.na(vs_names)]

    vs_names <- sort(vs_names)

    paste0(
      df_texts["locations_from", str_language],
      paste0(vs_names, collapse = "; ")
    )
    ## For testing purposes:
    # paste0(df_plotted_points)

  })


  ### "submit_deletion": process deletion request ####
  observeEvent(input$submit_deletion, {
    str_receided_del_cof <- trimws(input$deletion_confirm_code)
    if(str_receided_del_cof == str_deletion_verification_code)
    {

      unlink("./student_data/*.csv")

      createAlert(session = session,
                  anchorId = "deletion_confirmed",
                  title = "",
                  content = "The student data had been removed",
                  append = FALSE)


    } else
    {
      createAlert(session = session,
                  anchorId = "deletion_confirmed",
                  title = "",
                  content = "Sorry, the confirmation code is empty or not correct. Please try again.",
                  append = FALSE)
    }




  })


    ## Give the file an unique name
    str_data_file <- paste0(as.numeric(format(Sys.time(), "%OS3")) * 1000,
                            paste0(sample(c(LETTERS, letters), size=5,replace = TRUE),collapse = ""),
                            ".csv"
    )



  ### "submit_location": Submit student location ####
  observeEvent(input$submit_location, {

    ## Give the file an unique name
    str_data_file <- paste0(as.numeric(format(Sys.time(), "%OS3")) * 1000,
                            paste0(sample(c(LETTERS, letters), size=5,replace = TRUE),collapse = ""),
                            ".csv"
    )



    df_student <- data.frame(lon = input$click_coordinates[2],
                             lat = input$click_coordinates[1],
                             variable_of_interest = re_variable_of_interest(),
                             student_name = input$student_name,
                             time_of_submitting = Sys.time()
    )
    write.csv(x = df_student,
              file = paste0("./student_data/",str_data_file),
              row.names=FALSE)

    createAlert(session = session,
                anchorId = "thanks_for_submitting",
                title = "",
                content = df_texts["thanks_for_submitting", str_language],
                append = FALSE)

  })




  #### DEFINE OUTPUT user interfaces ####
  output$myUI <- renderUI({

    #browser()

    if(re_is_teacher_version())
    {  #### teacher version UI: map, histogram and test  ####

      ## Show locations on map
      ## Show statistics

      tagList(

        fluidRow(
          column(6,
                 leafletOutput(outputId = 'map', width = "100%", height = "80vh")
          ),
          column(6,
                 plotOutput(outputId = "histogram_of_var", width = "100%", height = "80vh")
          ),
        ),

        br(),

        textOutput("show_version"),
        #verbatimTextOutput("show_version"),
      )


    } else    {
      #### remove UI ####
      if(re_remove_student_data())
      {




        tagList(br(),

          textInput(inputId = "deletion_confirm_code",
                    label = "Please provide deletion confirmation code",
                    value = "",
                    width = NULL,
                    placeholder = NULL),

          actionButton(inputId = "submit_deletion",
                       label = "Delete student locations"),
          bsAlert("deletion_confirmed")
        )
        # tagList(br(),
        #         # hr(),
        #         # p("All student data has been removed"),
        #         # hr(),
        #         p()
        # )
      } else {
        #### student version UI: clickable map, webform ####
        tagList(

          leafletOutput(outputId = 'map',  height =  "65vh"), #width = "100%",

          br(),

          verbatimTextOutput("location_and_var"),

		 conditionalPanel(
			  condition = "input.click_coordinates != null" ,
			  fluidRow(
				column(3,
					   textInput(inputId = "student_name",
								 label = df_texts["ask_for_name", str_language],
								 value = "",
								 width = NULL,
								 placeholder = NULL),
				),

				column(3,
					   br(),

						actionButton(inputId = "submit_location",
									label = df_texts["submit_location", str_language],
									icon = icon("arrow-right"),
									width = NULL),

				),
				column(6,
					   bsAlert("thanks_for_submitting")
				),
			  ) ## End fluid row
		  ) ## End coditional panel
        )
      }
    }






  })



}

